@@ -44,6 +44,7 @@
 #include "kernel-lib/list_sort.h"
 #include "common/help.h"
 #include "common/rbtree-utils.h"
+#include "common/parse-utils.h"
 #include "mkfs/common.h"
 #include "mkfs/rootdir.h"
 #include "common/fsfeatures.h"
@@ -51,8 +52,6 @@
 #include "common/units.h"
 #include "check/qgroup-verify.h"
 
-static int verbose = 1;
-
 struct mkfs_allocation {
 	u64 data;
 	u64 metadata;
@@ -67,16 +66,13 @@ static int create_metadata_block_groups(struct btrfs_root *root, int mixed,
 	struct btrfs_trans_handle *trans;
 	struct btrfs_space_info *sinfo;
 	u64 flags = BTRFS_BLOCK_GROUP_METADATA;
-	u64 bytes_used;
 	u64 chunk_start = 0;
 	u64 chunk_size = 0;
-	u64 system_group_offset = BTRFS_BLOCK_RESERVED_1M_FOR_SUPER;
 	u64 system_group_size = BTRFS_MKFS_SYSTEM_GROUP_SIZE;
 	int ret;
 
 	if (btrfs_is_zoned(fs_info)) {
 		/* Two zones are reserved for superblock */
-		system_group_offset = fs_info->zone_size * BTRFS_NR_SB_LOG_ZONES;
 		system_group_size = fs_info->zone_size;
 	}
 
@@ -90,16 +86,12 @@ static int create_metadata_block_groups(struct btrfs_root *root, int mixed,
 
 	trans = btrfs_start_transaction(root, 1);
 	BUG_ON(IS_ERR(trans));
-	bytes_used = btrfs_super_bytes_used(fs_info->super_copy);
 
 	root->fs_info->system_allocs = 1;
 	/*
-	 * First temporary system chunk must match the chunk layout
-	 * created in make_btrfs().
+	 * We already created the block group item for our temporary system
+	 * chunk in make_btrfs(), so account for the size here.
 	 */
-	ret = btrfs_make_block_group(trans, fs_info, bytes_used,
-				     BTRFS_BLOCK_GROUP_SYSTEM,
-				     system_group_offset, system_group_size);
 	allocation->system += system_group_size;
 	if (ret)
 		return ret;
@@ -142,7 +134,6 @@ static int create_metadata_block_groups(struct btrfs_root *root, int mixed,
 
 	root->fs_info->system_allocs = 0;
 	ret = btrfs_commit_transaction(trans, root);
-
 err:
 	return ret;
 }
@@ -259,7 +250,11 @@ static int recow_roots(struct btrfs_trans_handle *trans,
 	ret = __recow_root(trans, info->csum_root);
 	if (ret)
 		return ret;
-
+	if (info->free_space_root) {
+		ret = __recow_root(trans, info->free_space_root);
+		if (ret)
+			return ret;
+	}
 	return 0;
 }
 
@@ -366,6 +361,7 @@ static void print_usage(int ret)
 	printf("\t-f|--force                  force overwrite of existing filesystem\n");
 	printf("  general:\n");
 	printf("\t-q|--quiet                  no messages except errors\n");
+	printf("\t-v|--verbose                increase verbosity level, default is 1\n");
 	printf("\t-V|--version                print the mkfs.btrfs version and exit\n");
 	printf("\t--help                      print this help and exit\n");
 	printf("  deprecated:\n");
@@ -375,30 +371,16 @@ static void print_usage(int ret)
 
 static u64 parse_profile(const char *s)
 {
-	if (strcasecmp(s, "raid0") == 0) {
-		return BTRFS_BLOCK_GROUP_RAID0;
-	} else if (strcasecmp(s, "raid1") == 0) {
-		return BTRFS_BLOCK_GROUP_RAID1;
-	} else if (strcasecmp(s, "raid1c3") == 0) {
-		return BTRFS_BLOCK_GROUP_RAID1C3;
-	} else if (strcasecmp(s, "raid1c4") == 0) {
-		return BTRFS_BLOCK_GROUP_RAID1C4;
-	} else if (strcasecmp(s, "raid5") == 0) {
-		return BTRFS_BLOCK_GROUP_RAID5;
-	} else if (strcasecmp(s, "raid6") == 0) {
-		return BTRFS_BLOCK_GROUP_RAID6;
-	} else if (strcasecmp(s, "raid10") == 0) {
-		return BTRFS_BLOCK_GROUP_RAID10;
-	} else if (strcasecmp(s, "dup") == 0) {
-		return BTRFS_BLOCK_GROUP_DUP;
-	} else if (strcasecmp(s, "single") == 0) {
-		return 0;
-	} else {
+	int ret;
+	u64 flags = 0;
+
+	ret = parse_bg_profile(s, &flags);
+	if (ret) {
 		error("unknown profile %s", s);
 		exit(1);
 	}
-	/* not reached */
-	return 0;
+
+	return flags;
 }
 
 static char *parse_label(const char *input)
@@ -927,6 +909,7 @@ int BOX_MAIN(mkfs)(int argc, char **argv)
 	u64 system_group_size;
 
 	crc32c_optimization_init();
+	btrfs_config_init();
 
 	while(1) {
 		int c;
@@ -952,12 +935,13 @@ int BOX_MAIN(mkfs)(int argc, char **argv)
 			{ "runtime-features", required_argument, NULL, 'R' },
 			{ "uuid", required_argument, NULL, 'U' },
 			{ "quiet", 0, NULL, 'q' },
+			{ "verbose", 0, NULL, 'v' },
 			{ "shrink", no_argument, NULL, GETOPT_VAL_SHRINK },
 			{ "help", no_argument, NULL, GETOPT_VAL_HELP },
 			{ NULL, 0, NULL, 0}
 		};
 
-		c = getopt_long(argc, argv, "A:b:fl:n:s:m:d:L:R:O:r:U:VMKq",
+		c = getopt_long(argc, argv, "A:b:fl:n:s:m:d:L:R:O:r:U:VvMKq",
 				long_options, NULL);
 		if (c < 0)
 			break;
@@ -1030,6 +1014,9 @@ int BOX_MAIN(mkfs)(int argc, char **argv)
 				block_count = parse_size_from_string(optarg);
 				zero_end = 0;
 				break;
+			case 'v':
+				bconf_be_verbose();
+				break;
 			case 'V':
 				printf("mkfs.btrfs, part of %s\n",
 						PACKAGE_STRING);
@@ -1046,7 +1033,7 @@ int BOX_MAIN(mkfs)(int argc, char **argv)
 				discard = 0;
 				break;
 			case 'q':
-				verbose = 0;
+				bconf_be_quiet();
 				break;
 			case GETOPT_VAL_SHRINK:
 				shrink_rootdir = true;
@@ -1060,7 +1047,7 @@ int BOX_MAIN(mkfs)(int argc, char **argv)
 		}
 	}
 
-	if (verbose) {
+	if (bconf.verbose) {
 		printf("%s\n", PACKAGE_STRING);
 		printf("See %s for more information.\n\n", PACKAGE_URL);
 	}
@@ -1127,7 +1114,7 @@ int BOX_MAIN(mkfs)(int argc, char **argv)
 			exit(1);
 		}
 	} else if (zoned_model(file) == ZONED_HOST_MANAGED) {
-		if (verbose)
+		if (bconf.verbose)
 			printf(
 	"Zoned: %s: host-managed device detected, setting zoned feature\n",
 			       file);
@@ -1143,7 +1130,7 @@ int BOX_MAIN(mkfs)(int argc, char **argv)
 		u64 tmp;
 
 		if (!metadata_profile_opt) {
-			if (dev_cnt == 1 && ssd && verbose)
+			if (dev_cnt == 1 && ssd && bconf.verbose)
 				printf("Detected a SSD, turning off metadata "
 				"duplication.  Mkfs with -m dup if you want to "
 				"force metadata duplication.\n");
@@ -1187,8 +1174,7 @@ int BOX_MAIN(mkfs)(int argc, char **argv)
 	if (mixed)
 		features |= BTRFS_FEATURE_INCOMPAT_MIXED_GROUPS;
 
-	if ((data_profile | metadata_profile) &
-	    (BTRFS_BLOCK_GROUP_RAID5 | BTRFS_BLOCK_GROUP_RAID6)) {
+	if ((data_profile | metadata_profile) & BTRFS_BLOCK_GROUP_RAID56_MASK) {
 		features |= BTRFS_FEATURE_INCOMPAT_RAID56;
 		warning("RAID5/6 support has known problems is strongly discouraged\n"
 			"\t to be used besides testing or evaluation.\n");
@@ -1321,7 +1307,7 @@ int BOX_MAIN(mkfs)(int argc, char **argv)
 
 	if (zoned && ((metadata_profile | data_profile) &
 		      BTRFS_BLOCK_GROUP_PROFILE_MASK)) {
-		error("cannot use RAID/DUP profile in zoned mode");
+		error("zoned mode does not yet support RAID/DUP profiles, please specify '-d single -m single' manually");
 		goto error;
 	}
 
@@ -1340,7 +1326,7 @@ int BOX_MAIN(mkfs)(int argc, char **argv)
 	ret = btrfs_prepare_device(fd, file, &dev_block_count, block_count,
 			(zero_end ? PREP_DEVICE_ZERO_END : 0) |
 			(discard ? PREP_DEVICE_DISCARD : 0) |
-			(verbose ? PREP_DEVICE_VERBOSE : 0) |
+			(bconf.verbose ? PREP_DEVICE_VERBOSE : 0) |
 			(zoned ? PREP_DEVICE_ZONED : 0));
 	if (ret)
 		goto error;
@@ -1359,10 +1345,9 @@ int BOX_MAIN(mkfs)(int argc, char **argv)
 		goto error;
 	}
 
-	if (group_profile_max_safe_loss(metadata_profile) <
-		group_profile_max_safe_loss(data_profile)){
+	if (btrfs_bg_type_to_tolerated_failures(metadata_profile) <
+	    btrfs_bg_type_to_tolerated_failures(data_profile))
 		warning("metadata has lower redundancy than data!\n");
-	}
 
 	mkfs_cfg.label = label;
 	memcpy(mkfs_cfg.fs_uuid, fs_uuid, sizeof(mkfs_cfg.fs_uuid));
@@ -1371,6 +1356,7 @@ int BOX_MAIN(mkfs)(int argc, char **argv)
 	mkfs_cfg.sectorsize = sectorsize;
 	mkfs_cfg.stripesize = stripesize;
 	mkfs_cfg.features = features;
+	mkfs_cfg.runtime_features = runtime_features;
 	mkfs_cfg.csum_type = csum_type;
 	mkfs_cfg.zone_size = zone_size(file);
 
@@ -1454,7 +1440,7 @@ int BOX_MAIN(mkfs)(int argc, char **argv)
 		}
 		ret = btrfs_prepare_device(fd, file, &dev_block_count,
 				block_count,
-				(verbose ? PREP_DEVICE_VERBOSE : 0) |
+				(bconf.verbose ? PREP_DEVICE_VERBOSE : 0) |
 				(zero_end ? PREP_DEVICE_ZERO_END : 0) |
 				(discard ? PREP_DEVICE_DISCARD : 0) |
 				(zoned ? PREP_DEVICE_ZONED : 0));
@@ -1468,7 +1454,7 @@ int BOX_MAIN(mkfs)(int argc, char **argv)
 			error("unable to add %s to filesystem: %d", file, ret);
 			goto error;
 		}
-		if (verbose >= 2) {
+		if (bconf.verbose >= 2) {
 			struct btrfs_device *device;
 
 			device = container_of(fs_info->fs_devices->devices.next,
@@ -1511,7 +1497,7 @@ raid_groups:
 	}
 
 	if (source_dir_set) {
-		ret = btrfs_mkfs_fill_dir(source_dir, root, verbose);
+		ret = btrfs_mkfs_fill_dir(source_dir, root, bconf.verbose);
 		if (ret) {
 			error("error while filling filesystem: %d", ret);
 			goto out;
@@ -1534,14 +1520,7 @@ raid_groups:
 			goto out;
 		}
 	}
-	if (runtime_features & BTRFS_RUNTIME_FEATURE_FREE_SPACE_TREE) {
-		ret = btrfs_create_free_space_tree(fs_info);
-		if (ret < 0) {
-			error("failed to create free space tree: %d (%m)", ret);
-			goto out;
-		}
-	}
-	if (verbose) {
+	if (bconf.verbose) {
 		char features_buf[64];
 
 		update_chunk_allocation(fs_info, &allocation);
